"""This module contains the wrapper class."""

import ctypes
from collections.abc import Callable
from pathlib import Path


class Wrapper:
    """A class that contains wrap functions."""

    def __init__(self: "Wrapper", path: Path) -> None:
        """Initializes the .so library.

        Args:
            path: The path to the .so library.

        Raises:
            OSError: If the .so library could not be initialized.
        """
        try:
            self.lib = ctypes.CDLL(f"{path}")
        except OSError as e:
            err_msg = f"Failed to initialize .so library: {path}. Error: {e!s}"
            raise OSError(err_msg) from e

    def __getattr__(self: "Wrapper", func: str) -> Callable[..., object]:
        """Gets the function from the library.

        Args:
            func: The name of the function.

        Returns:
            The function.

        Raises:
            AttributeError: If the function could not be found.
        """
        try:
            if isinstance(getattr(self.lib, func), ctypes._CFuncPtr):  # type: ignore[attr] # noqa: SLF001
                return getattr(self.lib, func)
        except AttributeError as e:
            err_msg = f"Failed to get function: {func}. Error: {e!s}"
            raise AttributeError(err_msg) from e
        return self.__getattr__(func)

    def __dir__(self: "Wrapper") -> list[str]:
        """Gets the functions in the library.

        Args:
            None

        Returns:
            list[str]: The functions in the library.

        Raises:
            Exception: If the functions could not be retrieved.
        """
        try:
            return [
                x
                for x in self.lib.__dict__
                if isinstance(getattr(self.lib, x), ctypes._CFuncPtr)  # type: ignore[attr] # noqa: SLF001
            ]
        except AttributeError as e:
            err_msg = f"Failed to get functions: {e!s}"
            raise AttributeError(err_msg) from e



"""This package contains utility functions."""

from dalmatium.database import Database
from dalmatium.forger import Forger
from dalmatium.reader import Reader
from dalmatium.wrapper import Wrapper


class Utils:
    """A class that contains utility functions."""

    (
        db,
        forge,
        read,
        wrap,
    ) = (
        Database,
        Forger,
        Reader,
        Wrapper,
    )



"""This module contains the Reader class."""

from pathlib import Path


class Reader:
    """A class that contains read functions."""

    @staticmethod
    def file(path: Path) -> str:
        """Read a file.

        Args:
            path: The path to the file.
        """
        try:
            return path.read_text()
        except Exception as e:
            err_msg = f"Failed to read file {path}. Error: {e!s}"
            raise OSError(err_msg) from e



"""This module contains the Forger class."""

from pathlib import Path


class Forger:
    """A class that contains forge functions."""

    @staticmethod
    def dir(
        path: Path,
        mode: int = 511,
        *,
        parents: bool = True,
        exist_ok: bool = True,
    ) -> Path:
        """Forge a directory if it does not exist.

        Args:
            path: The path to the directory.
            mode: The mode of the directory.
            parents: Whether to create parent directories.
            exist_ok: Whether to raise an error if the directory exists.

        Returns:
            The path to the directory.

        Raises:
            OSError: If the directory could not be created.
        """
        try:
            path.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        except OSError as e:
            err_msg = f"Failed to forge directory: {path}. Error: {e!s}"
            raise OSError(err_msg) from e
        return path

    @staticmethod
    def file(
        path: Path,
        mode: int = 438,
        *,
        exist_ok: bool = True,
    ) -> Path:
        """Forge a file if it does not exist.

        Args:
            path: The path to the file.
            mode: The mode of the file.
            exist_ok: Whether to raise an error if the file exists.

        Returns:
            The path to the file.

        Raises:
            OSError: If the file could not be created.
        """
        try:
            path.touch(mode=mode, exist_ok=exist_ok)
        except OSError as e:
            err_msg = f"Failed to forge file: {path}. Error: {e!s}"
            raise OSError(err_msg) from e
        return path



"""This module contains the database functionality."""

from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import sqlalchemy as sql
import sqlalchemy.orm as sqlo
from sqlalchemy.orm import Session


class Database:
    """A class that contains database functions."""

    base = sqlo.declarative_base()

    def __init__(self: "Database", path: Path) -> None:
        """Initialize the class.

        Args:
            path: The path to the database file.
        """
        self.__engine__ = sql.create_engine(str(f"sqlite:///{path}"))
        self.base.metadata.create_all(self.__engine__)
        self.__session__ = sqlo.sessionmaker(bind=self.__engine__)

    @contextmanager
    def db_session(self: "Database") -> Iterator[Session]:
        """Context manager for a database session.

        Returns:
            The database session.
        """
        session = self.__session__()
        try:
            yield session
            session.commit()
        except Exception:
            session.rollback()
            raise
        finally:
            session.close()

    def set_data(
        self: "Database",
        table: sqlo.DeclarativeMeta,
        data: dict[str, None | int | float | str | bytes],
    ) -> None:
        """Update data in the database or add if it does not exist.

        Args:
            table: The table to update the data in.
            data: The data to be updated.
        """
        with self.db_session() as session:
            for key, value in data.items():
                record = session.query(table).filter_by(key=key).first()
                if record is not None:
                    record.value = value
                else:
                    new_record = table(key=key, value=value)
                    session.add(new_record)

    def delete_data(
        self: "Database",
        table: sqlo.DeclarativeMeta,
        key: str,
    ) -> None:
        """Delete data from the database.

        Args:
            table: The table to delete the data from.
            key: The key of the data to be deleted.
        """
        with self.db_session() as session:
            record = session.query(table).filter_by(key=key).first()
            if record is not None:
                session.delete(record)

    def get_data(
        self: "Database",
        table: sqlo.DeclarativeMeta,
        key: str,
        default: None | float | str | bytes = "NULL",
    ) -> None | int | float | str | bytes:
        """Get data from the database.

        Args:
            table: The table to get the data from.
            key: The key to get the data for.
            default: The default value if the data does not exist.

        Returns:
            The data from the database.
        """
        with self.db_session() as session:
            record = session.query(table).filter_by(key=key).first()
            return record.value if record is not None else default



"""This module contains unit tests for the Wrapper class."""

import ctypes
import unittest

from dalmatium import Utils
from dalmatium.__config__ import UTILS_LIB_FILE


class TestWrapper(unittest.TestCase):
    """A class to test the Wrapper class."""

    def setUp(self: "TestWrapper") -> None:
        """Sets up the test."""
        self.wrapper = Utils.wrap(UTILS_LIB_FILE)
        self.arr = list(range(1, 6))
        self.arr_type = ctypes.c_int * len(self.arr)
        self.arr_c = self.arr_type(*self.arr)

    def test_factorial(self: "TestWrapper") -> None:
        """Tests the factorial function."""
        expected_result = 120
        if self.wrapper.factorial(5) != expected_result:
            raise AssertionError

    def test_sumarray(self: "TestWrapper") -> None:
        """Tests the sumArray function."""
        expected_result = 15
        if self.wrapper.sumArray(self.arr_c, len(self.arr)) != expected_result:
            raise AssertionError

    def test_ispalindrome(self: "TestWrapper") -> None:
        """Tests the isPalindrome function."""
        if not self.wrapper.isPalindrome(b"racecar"):
            raise AssertionError

    def test_isprime(self: "TestWrapper") -> None:
        """Tests the isPrime function."""
        if not self.wrapper.isPrime(7):
            raise AssertionError

    def test_maxarray(self: "TestWrapper") -> None:
        """Tests the maxArray function."""
        expected_result = 5
        if self.wrapper.maxArray(self.arr_c, len(self.arr)) != expected_result:
            raise AssertionError

    def test_minarray(self: "TestWrapper") -> None:
        """Tests the minArray function."""
        expected_result = 1
        if self.wrapper.minArray(self.arr_c, len(self.arr)) != expected_result:
            raise AssertionError



"""This module contains unit tests for the Forger class."""

import unittest
from pathlib import Path
from unittest.mock import MagicMock

from dalmatium.forger import Forger


class TestForger(unittest.TestCase):
    """A class to test the Forger class."""

    def setUp(self: "TestForger") -> None:
        """Sets up the test."""
        self.path = MagicMock(spec=Path)

    def test_forge_dir(self: "TestForger") -> None:
        """Tests the dir method."""
        Forger.dir(self.path)

        self.path.mkdir.assert_called_once_with(
            mode=511,
            parents=True,
            exist_ok=True,
        )

    def test_forge_file(self: "TestForger") -> None:
        """Tests the file method."""
        Forger.file(self.path)

        self.path.touch.assert_called_once_with(mode=438, exist_ok=True)



"""This module contains the setup configuration for Dalmatium."""

from pathlib import Path

from setuptools import find_packages, setup

setup(
    name="Dalmatium",
    version="1.0.0",
    url="https://github.com/IllyrionSoftware/Dalmatium",
    license=Path("LICENSE.md").read_text(),
    description="CAD addon/plug-in to automize frameworks.",
    long_description=Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    author="IllyrionSoftware",
    author_email="info@illyrion.eu",
    packages=find_packages(),
    install_requires=Path("requirements.txt").read_text().splitlines(),
    python_requires=">=3.12",
    package_data={"*": ["*.py", "py.typed"]},
    data_files=[
        (
            "",
            [
                "CODE_OF_CONDUCT.md",
                "CONTRIBUTING.md",
                "LICENSE.md",
                "README.md",
                "requirements.txt",
            ],
        ),
    ],
    zip_safe=False,
    project_urls={
        "Source Code": "https://github.com/IllyrionSoftware/Dalmatium",
        "Bug Tracker": "https://github.com/IllyrionSoftware/Dalmatium/issues",
        "Documentation": "https://wiki.illyrion.eu/",
        "Funding": "https://buymeacoffee.com/illyrius",
    },
)



"""This module contains unit tests for the Database class."""

import unittest
from pathlib import Path

from sqlalchemy import Column, String

from dalmatium import Utils
from dalmatium.__config__ import TEST_DATABASE_FILE
from dalmatium.database import Database


class TestTable(Utils.db.base):
    """A class to represent a test table."""

    __tablename__ = "test_table"

    key = Column(String(50), primary_key=True)
    value = Column(String(50))


class TestDatabase(unittest.TestCase):
    """A class to test the Database class."""

    @classmethod
    def setUpClass(cls: type["TestDatabase"]) -> None:
        """Sets up the test."""
        cls.db_path: Path = Path(TEST_DATABASE_FILE)
        cls.db: Database = Database(cls.db_path)
        cls.TestTable = TestTable
        cls.db.base.metadata.create_all(cls.db.__engine__)
        cls.db.base.metadata.reflect(bind=cls.db.__engine__)

    def test_set_data(self: "TestDatabase") -> None:
        """Tests the set_data method."""
        self.db.set_data(
            self.TestTable,
            {
                "key1": "value1",
                "key2": "value2",
            },
        )
        with self.db.db_session() as session:
            record1 = (
                session.query(self.TestTable).filter_by(key="key1").first()
            )
            record2 = (
                session.query(self.TestTable).filter_by(key="key2").first()
            )
            if record1 is None:
                raise AssertionError
            if record2 is None:
                raise AssertionError

    def test_delete_data(self: "TestDatabase") -> None:
        """Tests the delete_data method."""
        key = "key1"
        self.db.set_data(
            self.TestTable,
            {
                "key1": "value1",
                "key2": "value2",
            },
        )
        self.db.delete_data(self.TestTable, key)
        with self.db.db_session() as session:
            record = session.query(self.TestTable).filter_by(key=key).first()
            if record is not None:
                raise AssertionError

    def test_get_data(self: "TestDatabase") -> None:
        """Tests the get_data method."""
        key, value = "key1", "value1"
        self.db.set_data(self.TestTable, {key: value})
        result = self.db.get_data(self.TestTable, key)
        if result != value:
            raise AssertionError


if __name__ == "__main__":
    unittest.main()
